#include <assert.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <glib.h>
#include <glib-object.h>
#include <dbus/dbus-glib.h>
#include <dbus/dbus-glib-lowlevel.h>
#include <xtalk/debug.h>
#include <xtalk/xusb.h>
#include <xpanel/xpw.h>
#include <xpanel/xpw_menu.h>
#include <xpanel/xpw_dialog.h>
#include <xpanel/xpw_ipaddr.h>
#include <xpanel/xpw_editbox.h>
#include <xpanel/xpw_displaybox.h>
#include <xpanel/xpw_passwd.h>	/* for XPW_PASSWD_NO_FACADE */
#include <xpanel/pcptalk.h>
#include <xpanel/kbd_glib.h>
#include <xpanel-ui.h>
#include "xpanel-dbus.h"
#include "xpanel-auth.h"

static char	*progname;

#define	DBG_MASK	0x80

#define XP_SYSCHECK_TIMEOUT  120000  // System temperature check timeout (msec)
#define XP_PSUCHECK_TIMEOUT  1000   // Internal PSU status check timeout (msec)

#define MAX_LOGO_LEN 16
#define DEFAULT_LOGO "XORCOM"

static void usage()
{
	fprintf(stderr, "Usage:\n");
	fprintf(stderr, "%s [options...]\n", progname);
	fprintf(stderr, "\t\t[-f]               # Run in foreground\n");
	fprintf(stderr, "\t\t[-c <character>]   # Hide password with character (default: exposed)\n");
	fprintf(stderr, "\t\t[-t <seconds>]     # Seconds to activate screen lock (default: %d)\n", SCREENLOCK_TIME);
	fprintf(stderr, "\t\t[-u <name>]        # Login user name (default: xpanel)\n");
	fprintf(stderr, "\t\t[-l <text>]        # Logo (max length is 16, default: XORCOM)\n");
	fprintf(stderr, "\t\t[-p <seconds>]     # Internal PSU monitoring period (default: %d. Define 0 to disable monitoring.\n", XP_PSUCHECK_TIMEOUT/1000);
	fprintf(stderr, "\t\t[-v]               # Increase verbosity\n");
	fprintf(stderr, "\t\t[-d mask]          # Debug mask (0xFF for everything)\n");
	exit(1);
}

#define EP_OUT  0x04
#define EP_IN   0x88

static struct xusb_spec	xusb_spec1;

static GMainLoop		*mainloop;
static GMainContext		*mainctx;
static struct xpw_screen	*screen;

static int xpanel_kbd_notification(struct xpanel_kbd *kbd)
{
	struct xpw		*widget;
	enum keycodes		key;
	int			i;

	INFO("%s: %d keys\n", __FUNCTION__, kbd->numkeys);
	
	widget = xpw_current();
	for(i = 0; i < kbd->numkeys; i++) {
		struct xpw	*next;
		key = kbd->keybuf[i];
		xpw_screen_beep(screen, 40000);
		stop_flashing();
		widget_redraw(screen, widget);
		next = process_key(widget, key);
		INFO("%s: \"%s\": \"%s\" -> \"%s\"\n", __FUNCTION__,
			keycode_name(key), widget->name, next->name);
		widget = next;
	}
	widget_redraw(screen, widget);
	kbd->numkeys = 0;
	return 0;
}

static gboolean sysTemperature_timeout_cb(gpointer data)
{
	struct pcp_device	*pcpdev = (struct pcp_device *)data;
	guchar				fan0_speed =0,
						fan1_speed =0,
						fan2_speed =0;
	gboolean			rc = FALSE;
	
	// Get current system temperature
	rc = get_fan_speed_ui(&fan0_speed, &fan1_speed, &fan2_speed);
	if(rc) {
		pcp_fanspeed_set(pcpdev, 0, fan0_speed);
		pcp_fanspeed_set(pcpdev, 1, fan1_speed);
		pcp_fanspeed_set(pcpdev, 2, fan2_speed);
	}
	
	return TRUE;
}

int run_mainloop(XPanelDaemon *daemon, struct xpanel_ui *ui)
{
	struct xpanel_kbd	*kbd;
	struct pcp_device	*pcpdev = ui->pcpdev;
	int					rc = 0;
	int					psu_sts = PCP_PSU_STATUS_DISCONNECTED;


        fprintf(stderr, "Starting\n");
        if((mainctx = g_main_context_default()) == NULL) {
                fprintf(stderr, "main context default\n");
                return 1;
        }
        if((mainloop = g_main_loop_new(mainctx, FALSE)) == NULL) {
                fprintf(stderr, "main loop\n");
                return 1;
        }
	if(!xpanel_daemon_dbus_register(daemon)) {
		g_critical ("Failed registering XPanelDaemon on the system bus\n");
		return 1;
	}
	fprintf(stderr, "Registered on dbus\n");
	// Add keyboard polling 
	kbd = kbd_glib_register(pcpdev, mainctx, xpanel_kbd_notification);
	if(kbd == NULL) {
		ERR("kbd_glib_register() failed\n");
		return 1;
	}
	kbd_idle_set_callback(kbd, xpanel_kbd_idle, ui);
	kbd_idle_set_timeout(kbd, ui->screen_lock_time);
	
	// Add system temperature polling. Needed for fan speed control.
	g_timeout_add(XP_SYSCHECK_TIMEOUT, sysTemperature_timeout_cb, (gpointer)pcpdev);

	//If protocol >0x10 then start PSU monitoring if necessary.
	if(pcp_get_panel_proto_version(pcpdev) > 0x10 && ui->psu_check_time > 0) {
		rc = pcp_psu_read(pcpdev, &psu_sts);
		//[FDO]
		INFO("%s: psu_sts=%d\n", __FUNCTION__, psu_sts);
		if(rc >= 0 && psu_sts != PCP_PSU_STATUS_DISCONNECTED)
			g_timeout_add(ui->psu_check_time, psu_check_timeout_cb, ui);
	}
	
	screen_lock(ui);	/* Start with a locked screen */
	kbd_flush(kbd);
        g_main_loop_run(mainloop);
        fprintf(stderr, "Ending\n");
	kbd_glib_destroy(kbd);
        return 0;
}

int main(int argc, char *argv[])
{
	struct xusb_iface	*xusb_iface;
	//const char		options[] = "fvc:d:EFpt:u:l:";
	const char		options[] = "fvc:d:p:t:u:l:";
	gboolean		run_in_foreground = FALSE;
	const struct xpw_class	*this_class;
	XPanelDaemon		*daemon1;
	struct pcp_device	*pcpdev;
	char			facade_char = XPW_PASSWD_NO_FACADE;
	int				lock_time = SCREENLOCK_TIME;
	int				psu_time = XP_PSUCHECK_TIMEOUT;
	char			*login_name = SCREENLOCK_USER;
	char			*logo = DEFAULT_LOGO;
	struct xpanel_ui	*ui;

	progname = argv[0];
	while (1) {
		int	c;

		c = getopt (argc, argv, options);
		if (c == -1)
			break;

		switch (c) {
			case 'v':
				verbose++;
				break;
			case 'd':
				debug_mask = strtoul(optarg, NULL, 0);
				break;
			case 'c':
				facade_char = optarg[0];
				break;
			case 'f':
				run_in_foreground = TRUE;
				break;
			case 'p':
				psu_time = strtoul(optarg, NULL, 0) * 1000;
				if (psu_time < 0) {
					ERR("Bad PSU check time: %d\n", psu_time);
					usage();
				}
				break;
			case 't':
				lock_time = strtoul(optarg, NULL, 0);
				if (lock_time <= 0) {
					ERR("Bad screen lock time: %d\n", lock_time);
					usage();
				}
				break;
			case 'u':
				login_name = strdup(optarg);
				if (!login_name) {
					ERR("Failed allocating memory\n");
					return 1;
				}
				break;
			case 'l':
				logo = strdup(optarg);
				if (!logo) {
					ERR("Failed allocating memory\n");
					return 1;
				if (strlen(logo) > MAX_LOGO_LEN)
					logo[MAX_LOGO_LEN] = 0;
				}
				break;
			case 'h':
			default:
				ERR("Unknown option '%c'\n", c);
				usage();
		}
	}

	if (!run_in_foreground)
		if (daemon(0,0) == -1) {
			ERR("Xpanel daemon initialization failure.\n");
			return 1;
		}
			
		
	#if !GLIB_CHECK_VERSION(2,35,0)
	g_type_init ();
	#endif
	xusb_init_spec(&xusb_spec1, "pcptalk", 0xe4e4, 0x1193);
	xusb_iface = xusb_open_one(&xusb_spec1, 1, 1, NULL, NULL);
	if (!xusb_iface) {
		ERR("Failed opening XUSB device\n");
		return 1;
	}
	pcpdev = pcp_new(xusb_iface);
	if(!pcpdev) {
		ERR("Failed calling pcp_new()\n");
		return 1;
	}
	pcp_lcd_function_set(pcpdev, 1, 1, 0);
	pcp_lcd_display_control(pcpdev, 1, 1, 1);
	pcp_lcd_clear_display(pcpdev);
	pcp_backlight(pcpdev, 1);
	if((screen = xpw_screen_new(pcpdev)) == NULL) {
		ERR("Failed creating xpw_screen\n");
		return 1;
	}
	INFO("Device and Protocol are ready (screen-lock-user='%s' (%s PAM)\n",
		login_name, (xpanel_auth_have_pam()) ? "with" : "without");
	if((ui = xpanel_ui_new(pcpdev, screen, lock_time, psu_time, login_name, logo)) == NULL) {
		ERR("Failed initializing User Interface\n");
		return 1;
	}
	ui->screen_lock_facade = facade_char;
	this_class = xpw_myclass(ui->toplevel);
	if(!this_class) {
		ERR("Failed getting class for classid=%d\n", ui->toplevel->classid);
		return 1;
	}
	this_class->dump_widget(ui->toplevel, 0, "");
	if((daemon1 = xpanel_daemon_new(screen, xpw_background_get(), raid_sts_cb)) == NULL) {
		g_critical ("Failed creating XPanelDaemon instance\n");
		return 1;
	}
	xpanel_ui_bind(ui, daemon1);
	fprintf(stderr, "Got new XPanelDaemon\n");
	run_mainloop(daemon1, ui);
	xpanel_ui_destroy(ui);
	xpw_screen_destroy(screen);
	pcp_delete(pcpdev);
	return 0;
}
