#include <assert.h>
#include <errno.h>
#include <stdio.h>
#include <string.h>
#include <glib.h>
#include <glib-object.h>
#include <dbus/dbus-glib.h>
#include <dbus/dbus-glib-lowlevel.h>
#include <xtalk/debug.h>
#include <xpanel/xpw_background.h>
#include "xpanel-dbus.h"
#include "xpanel-dbus-public.h"

/*
 * Our DBUS methods
 */
static gboolean xpanel_backlight (XPanelDaemon *daemon, gboolean on, DBusGMethodInvocation *context);
static gboolean xpanel_background_text (XPanelDaemon *daemon, gchar *text, DBusGMethodInvocation *context);
static gboolean xpanel_buzzer (XPanelDaemon *daemon, gboolean on, DBusGMethodInvocation *context);
static gboolean xpanel_raid_info (XPanelDaemon *daemon, gchar *text, DBusGMethodInvocation *context);
static gint xpanel_fanspeed (XPanelDaemon *daemon, gint fan_number, gint delta, DBusGMethodInvocation *context);
static gboolean xpanel_zigbee_lamp (XPanelDaemon *daemon, gint lamp_num, gboolean on, DBusGMethodInvocation *context);

#include "xpanel-dbus-glue.h"
#include "xpanelbe-dbus-glue.h"
#include "xrps-dbus-glue.h"
#include "xrps-manager-dbus-glue.h"

#define	XPANEL_SERVICE_NAME	"com.xorcom.XPanel"
#define	XPANEL_VERSION		"1.0"

#define	XPANELBE_SERVICE_NAME		"com.xorcom.XPanelBE"
#define	XPANELBE_SERVICE_OBJECT_PATH	"/com/xorcom/XPanelBE"
#define	XPANELBE_SERVICE_INTERFACE	"com.xorcom.XPanelBE"

#define	MPS_SERVICE_NAME		"com.xorcom.xrps"
#define	MPS_SERVICE_OBJECT_PATH	"/com/xorcom/xrps/"
#define	MPS_SERVICE_INTERFACE	"com.xorcom.xrps"

#define	MPS_MANAGER_SERVICE_OBJECT_PATH	"/com/xorcom/xrps/Manager"
#define	MPS_MANAGER_SERVICE_INTERFACE	"com.xorcom.xrps.Manager"

G_DEFINE_TYPE (XPanelDaemon, xpanel_daemon, G_TYPE_OBJECT);

#define XPANEL_DAEMON_GET_PRIVATE(o) (G_TYPE_INSTANCE_GET_PRIVATE ((o), XPANEL_TYPE_DAEMON, XPanelDaemonPrivate))

/*
 * Our DBUS properties
 */
enum {
	PROP_0,
	PROP_VERSION,
	PROP_LAST
};

/*
 * Our DBUS signals
 */
enum
{
	SIGNAL_BACKLIGHT_CHANGED,
	SIGNAL_NETWORK_PARAMETERS,
	SIGNAL_LAST,
};

static guint		signals[SIGNAL_LAST] = { 0 };

struct XPanelDaemonPrivate {
	DBusGConnection		*connection;
	DBusGProxy		*proxy;
	DBusGProxy		*xpanelbe_proxy;
	DBusGProxy		*mps_proxy;
	DBusGProxy		*mps_manager_proxy;
	struct xpw_screen	*screen;
	struct xpw_background	*background;
	gchar			*version;
	raid_sts_cb_t	raid_sts_cb;
};

static void xpanel_daemon_finalize (GObject *object)
{
	XPanelDaemon	*daemon;

	fprintf(stderr, "%s\n", __func__);
	g_return_if_fail (object != NULL);
	g_return_if_fail (XPANEL_IS_DAEMON (object));
	daemon = XPANEL_DAEMON (object);
	if(daemon->priv->xpanelbe_proxy != NULL)
		g_object_unref (daemon->priv->xpanelbe_proxy);
	if(daemon->priv->mps_proxy != NULL)
		g_object_unref (daemon->priv->mps_proxy);
	g_return_if_fail (daemon->priv != NULL);
	if (daemon->priv->proxy != NULL)
		g_object_unref (daemon->priv->proxy);
	if (daemon->priv->connection != NULL)
		dbus_g_connection_unref (daemon->priv->connection);
	if (daemon->priv->version != NULL)
		g_free (daemon->priv->version);
	G_OBJECT_CLASS (xpanel_daemon_parent_class)->finalize (object);
}

XPanelDaemon *xpanel_daemon_new(struct xpw_screen *screen, struct xpw_background *background, raid_sts_cb_t raid_sts_cb)
{
	XPanelDaemon	*daemon;

	daemon = XPANEL_DAEMON (g_object_new (XPANEL_TYPE_DAEMON, NULL));
	daemon->priv->screen = screen;
	daemon->priv->background = background;
	daemon->priv->raid_sts_cb = raid_sts_cb;
	return daemon;
}

static void xpanel_daemon_get_property (GObject *object, guint prop_id, GValue *value, GParamSpec *pspec)
{
	XPanelDaemon	*daemon = XPANEL_DAEMON (object);

	fprintf(stderr, "%s\n", __func__);
	switch (prop_id) {
	case PROP_VERSION:
		g_value_set_string (value, daemon->priv->version);
		break;
	default:
		G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
		break;
	}
}

static void xpanel_daemon_set_property()
{
	fprintf(stderr, "%s\n", __func__);
}

static void xpanel_daemon_init (XPanelDaemon *daemon)
{
	fprintf(stderr, "%s\n", __func__);
	daemon->priv = XPANEL_DAEMON_GET_PRIVATE (daemon);
	daemon->priv->version = g_strdup (XPANEL_VERSION);
}

GQuark xpanel_daemon_error_quark (void)
{
	static GQuark ret = 0;
	if (ret == 0)
		ret = g_quark_from_static_string ("xpanel_daemon_error");
	return ret;
}

#define ENUM_ENTRY(NAME, DESC) { NAME, "" #NAME "", DESC }

GType xpanel_daemon_error_get_type (void)
{
	static GType etype = 0;

	if (etype == 0) {
		static const GEnumValue values[] = {
			ENUM_ENTRY (XPANEL_DAEMON_ERROR_GENERAL, "GeneralError"),
			ENUM_ENTRY (XPANEL_DAEMON_ERROR_NOT_SUPPORTED, "NotSupported"),
			ENUM_ENTRY (XPANEL_DAEMON_ERROR_NO_SUCH_DEVICE, "NoSuchDevice"),
			{ 0, 0, 0 }
		};
		g_assert (XPANEL_DAEMON_NUM_ERRORS == G_N_ELEMENTS (values) - 1);
		etype = g_enum_register_static ("XPanelDaemonError", values);
	}
	return etype;
}

static void xpanel_daemon_class_init (XPanelDaemonClass *klass)
{
	GObjectClass	*object_class = G_OBJECT_CLASS (klass);

	fprintf(stderr, "%s:\n", __func__);
	object_class->finalize = xpanel_daemon_finalize;
	object_class->get_property = xpanel_daemon_get_property;
	object_class->set_property = xpanel_daemon_set_property;
	g_type_class_add_private (klass, sizeof (XPanelDaemonPrivate));
	signals[SIGNAL_BACKLIGHT_CHANGED] =
		g_signal_new ("backlightchanged",
			      G_OBJECT_CLASS_TYPE (klass),
			      G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
			      0, NULL, NULL,
			      g_cclosure_marshal_VOID__STRING,
			      G_TYPE_NONE, 1, G_TYPE_STRING);
	fprintf(stderr, "%s: Signal %d registered\n", __func__, signals[SIGNAL_BACKLIGHT_CHANGED]);
	signals[SIGNAL_NETWORK_PARAMETERS] =
		g_signal_new ("network_parameters",
			      G_OBJECT_CLASS_TYPE (klass),
			      G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
			      0, NULL, NULL,
			      g_cclosure_marshal_VOID__STRING,
			      G_TYPE_NONE, 5, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRING);
	fprintf(stderr, "%s: Signal %d registered\n", __func__, signals[SIGNAL_NETWORK_PARAMETERS]);
	dbus_g_object_type_install_info (XPANEL_TYPE_DAEMON, &dbus_glib_xpanel_object_info);
	g_object_class_install_property (object_class,
					 PROP_VERSION,
					 g_param_spec_string ("version", NULL, NULL, NULL, G_PARAM_READABLE));
	dbus_g_error_domain_register (XPANEL_DAEMON_ERROR, NULL, XPANEL_DAEMON_TYPE_ERROR);
}

/*
 * Make us owners of com.xorcom.XPanel interface
 */
static gboolean xpanel_acquire_name(DBusGConnection *connection, const gchar *name)

{
	DBusGProxy	*bus_proxy;
	GError		*error = NULL;
	guint		result;
	gboolean	ret = FALSE;

	/* get proxy */
	bus_proxy = dbus_g_proxy_new_for_name (connection, DBUS_SERVICE_DBUS,
					       DBUS_PATH_DBUS, DBUS_INTERFACE_DBUS);
	if (bus_proxy == NULL) {
		g_critical ("Could not construct bus_proxy object; bailing out");
		goto out;
	}

	/* aquire name */
	ret = dbus_g_proxy_call (bus_proxy, "RequestName", &error,
				 G_TYPE_STRING, name,
				 G_TYPE_UINT, 0,
				 G_TYPE_INVALID,
				 G_TYPE_UINT, &result,
				 G_TYPE_INVALID);
	if (!ret) {
		if (error != NULL) {
			g_critical ("Failed to acquire %s: %s", name, error->message);
			g_error_free (error);
		} else {
			g_critical ("Failed to acquire %s", name);
		}
		goto out;
	}

	/* already taken */
 	if (result != DBUS_REQUEST_NAME_REPLY_PRIMARY_OWNER) {
		if (error != NULL) {
			g_critical ("Failed to acquire %s: %s", name, error->message);
			g_error_free (error);
		} else {
			g_critical ("Failed to acquire %s", name);
		}
		ret = FALSE;
		goto out;
	}
out:
	return ret;
}

gboolean xpanel_daemon_dbus_register (XPanelDaemon *daemon)
{
	GError		*error = NULL;
	gboolean	ret = FALSE;

	fprintf(stderr, "%s:\n", __func__);
	daemon->priv->connection = dbus_g_bus_get (DBUS_BUS_SYSTEM, &error);
	if (daemon->priv->connection == NULL) {
		if (error != NULL) {
			g_critical ("error getting system bus: %s", error->message);
			g_error_free (error);
		}
		g_critical ("No connection established");
		goto out;
	}
	fprintf(stderr, "%s: Got connection\n", __func__);

	/* connect to DBUS */
	daemon->priv->proxy = dbus_g_proxy_new_for_name (daemon->priv->connection,
							 DBUS_SERVICE_DBUS,
							 DBUS_PATH_DBUS,
							 DBUS_INTERFACE_DBUS);

	fprintf(stderr, "%s: Got proxy\n", __func__);
	/* register GObject */
	dbus_g_connection_register_g_object (daemon->priv->connection,
					     "/com/xorcom/XPanel",
					     G_OBJECT (daemon));
	if(!xpanel_acquire_name(daemon->priv->connection, XPANEL_SERVICE_NAME)) {
		g_critical ("Failed to acquire name for connection");
		goto out;
	}
	fprintf(stderr, "%s: Registered our object\n", __func__);
	/* success */
	ret = TRUE;
out:
	return ret;
}

static gboolean xpanel_backlight (XPanelDaemon *daemon, gboolean on, DBusGMethodInvocation *context)
{
	struct pcp_device	*pcpdev;
	const char		*msg = (on) ? "ON" : "OFF";

	fprintf(stderr, "%s: %s\n", __func__, msg);
	pcpdev = xpw_screen_getphys(daemon->priv->screen);
	pcp_backlight(pcpdev, on);
	g_signal_emit (daemon, signals[SIGNAL_BACKLIGHT_CHANGED], 0, msg);
	return 1;
}

static gboolean xpanel_background_text (XPanelDaemon *daemon, gchar *text, DBusGMethodInvocation *context)
{
	struct xpw_screen	*screen;
	struct xpw_background	*background;

	fprintf(stderr, "%s: %s\n", __func__, text);
	screen = daemon->priv->screen;
	background = daemon->priv->background;
	xpw_background_addstring(background, text);
	widget_redraw(screen, WIDGET(background));
	return 1;
}

static gboolean xpanel_buzzer (XPanelDaemon *daemon, gboolean on, DBusGMethodInvocation *context)
{
	struct pcp_device	*pcpdev;
	const char		*msg = (on) ? "ON" : "OFF";

	fprintf(stderr, "%s: %s\n", __func__, msg);
	pcpdev = xpw_screen_getphys(daemon->priv->screen);
	pcp_buzzer(pcpdev, on);
	return 1;
}

static gboolean xpanel_raid_info (XPanelDaemon *daemon, gchar *text, DBusGMethodInvocation *context)
{
	// This method may be called by an external application to inform about a RAID problem.
	// Here we call the raid_sts_cb() (ref. xpanel-ui.c) call-back function.
	// The raid_sts_cb() will take care to show a notify message on LCD in correct way.
	fprintf(stderr, "%s: %s\n", __func__, text);
	(*daemon->priv->raid_sts_cb)((const char*)text);
	return 1;
}

static gint xpanel_fanspeed (XPanelDaemon *daemon, gint fan_number, gint speed, DBusGMethodInvocation *context)
{
	struct pcp_device	*pcpdev;
	gint			prevspeed;

	fprintf(stderr, "%s: fan=%d speed=%d\n", __func__, fan_number, speed);
	pcpdev = xpw_screen_getphys(daemon->priv->screen);
	prevspeed = pcp_fanspeed_set(pcpdev, fan_number, speed);
	return prevspeed;
}

static gboolean xpanel_zigbee_lamp (XPanelDaemon *daemon, gint lamp_num, gboolean on, DBusGMethodInvocation *context)
{
	struct pcp_device	*pcpdev;

	fprintf(stderr, "%s: lamp=%d %s\n", __func__, lamp_num, (on) ? "ON" : "OFF");
	pcpdev = xpw_screen_getphys(daemon->priv->screen);
	pcp_zigbee_lamp(pcpdev, lamp_num, on);
	return 1;
}

int set_network_parameters(
	const char *str_ipaddr,
	const char *str_netmask,
	const char *str_defaultgw,
	const char *str_dns1,
	const char *str_dns2,
	void *data)
{
	XPanelDaemon	*daemon = data;

	fprintf(stderr, "User want IP=\"%s/%s\" GW=\"%s\" DNS1=\"%s\" DNS2=\"%s\"\n",
		str_ipaddr, str_netmask, str_defaultgw, str_dns1, str_dns2);
	g_signal_emit (daemon, signals[SIGNAL_NETWORK_PARAMETERS], 0,
		str_ipaddr, str_netmask, str_defaultgw, str_dns1, str_dns2);
	return 0;
}

static void xpanelbe_proxy_destroy(struct XPanelDaemonPrivate *xpanel_priv)
{
	assert(xpanel_priv);
	INFO("%s\n", __func__);
	if(xpanel_priv->xpanelbe_proxy != NULL) {
		g_object_unref(xpanel_priv->xpanelbe_proxy);
		xpanel_priv->xpanelbe_proxy = NULL;
	}
}

static DBusGProxy *xpanelbe_proxy_get(struct XPanelDaemonPrivate *xpanel_priv)
{
	DBusGProxy	*proxy;

	assert(xpanel_priv);
	INFO("%s\n", __func__);
	if(xpanel_priv->xpanelbe_proxy)
		return xpanel_priv->xpanelbe_proxy;
	proxy = dbus_g_proxy_new_for_name(xpanel_priv->connection,
                              XPANELBE_SERVICE_NAME,
                              XPANELBE_SERVICE_OBJECT_PATH,
                              XPANELBE_SERVICE_INTERFACE);
	if(!proxy) {
		ERR("Failed getting XPanelBE proxy\n");
		return NULL;
	}
	xpanel_priv->xpanelbe_proxy = proxy;
	return proxy;
}

static void mps_proxy_destroy(struct XPanelDaemonPrivate *xpanel_priv)
{
	assert(xpanel_priv);
	INFO("%s\n", __func__);
	if(xpanel_priv->mps_proxy != NULL) {
		g_object_unref (xpanel_priv->mps_proxy);
		xpanel_priv->mps_proxy = NULL;
	}
}

static DBusGProxy *mps_manager_proxy_get(struct XPanelDaemonPrivate *xpanel_priv)
{
	DBusGProxy	*proxy;

	assert(xpanel_priv);
	INFO("%s\n", __func__);
	if(xpanel_priv->mps_manager_proxy)
		return xpanel_priv->mps_manager_proxy;
	proxy = dbus_g_proxy_new_for_name(xpanel_priv->connection,
                              MPS_SERVICE_NAME,
                              MPS_MANAGER_SERVICE_OBJECT_PATH,
                              MPS_MANAGER_SERVICE_INTERFACE);
	if(!proxy) {
		ERR("Failed getting MPS Manager proxy\n");
		return NULL;
	}
	xpanel_priv->mps_manager_proxy = proxy;
	return proxy;
}

static DBusGProxy *mps_proxy_get(struct XPanelDaemonPrivate *xpanel_priv, char * device)
{
	DBusGProxy	*proxy;
	char		object_name[100];

	assert(xpanel_priv);
	INFO("%s\n", __func__);
	if(xpanel_priv->mps_proxy)
		return xpanel_priv->mps_proxy;
		
	object_name[0] = 0;	
	strcpy(object_name, MPS_SERVICE_OBJECT_PATH);
	strcat(object_name, device);
	proxy = dbus_g_proxy_new_for_name(xpanel_priv->connection,
                              MPS_SERVICE_NAME,
                              object_name,
                              MPS_SERVICE_INTERFACE);
	if(!proxy) {
		ERR("Failed getting MPS proxy\n");
		return NULL;
	}
	xpanel_priv->mps_proxy = proxy;
	return proxy;
}


GList *get_interface_names(void *data)
{
	XPanelDaemon		*daemon = data;
	GList			*list = NULL;
	DBusGProxy		*proxy;
	guchar			status = 77;
	char			*errormsg = NULL;
	char			**interfaces = NULL;
	GError			*error = NULL;
	gboolean		result;
	int			i;

	assert(daemon);
	INFO("%s\n", __func__);
	if((proxy = xpanelbe_proxy_get(daemon->priv)) == NULL)
		return NULL;
	result = com_xorcom_XPanelBE_list_interfaces (proxy, &status, &errormsg, &interfaces, &error);
	xpanelbe_proxy_destroy(daemon->priv);
	if(!result) {
		ERR("Failed in XPanelBE.ListInterfaces\n");
		return NULL;
	}
	INFO("%s: (%d) = status=%d errmsg='%s'\n", __func__, result, status, errormsg);
	for(i = 0; interfaces[i]; i++) {
		if(strncmp(interfaces[i], "lo", 2) == 0)
			continue;
		INFO("%s --> '%s'\n", __func__, interfaces[i]);
		list = g_list_append (list, strdup(interfaces[i]));
	}
	if( error != NULL)
		g_error_free(error);
	g_strfreev(interfaces);
	return list;
}

void free_interface_names(void *data, GList *list)
{
	XPanelDaemon	*daemon = data;

	assert(daemon);
	fprintf(stderr, "%s\n", __func__);
	if(!list)
		return;
	g_list_free(list);
}

int dhcp_setter(void *context, const char *netinterface)
{
	XPanelDaemon	*daemon = context;
	DBusGProxy	*proxy;
	guchar		status = 77;
	char		*errormsg = NULL;
	GError		*error = NULL;
	gboolean	result;

	INFO("%s(%s)\n", __func__, netinterface);
	if((proxy = xpanelbe_proxy_get(daemon->priv)) == NULL)
		return -EFAULT;
	result = com_xorcom_XPanelBE_set_dhcp_client (proxy, netinterface, &status, &errormsg, &error);
	xpanelbe_proxy_destroy(daemon->priv);
	if( error != NULL)
		g_error_free(error);
	g_free(errormsg);
	return (result) ? 0 : -EFAULT;
}

gboolean get_interface_settings(void *context, const char *netinterface, char **ipaddr, char **netmask)
{
	XPanelDaemon	*daemon = context;
	DBusGProxy	*proxy;
	guchar		status = 77;
	char		*errormsg = NULL;
	GError		*error = NULL;
	gboolean	result;

	INFO("%s(%s)\n", __func__, netinterface);
	if((proxy = xpanelbe_proxy_get(daemon->priv)) == NULL)
		return FALSE;
	result = com_xorcom_XPanelBE_interface_settings(proxy, netinterface, &status, &errormsg, ipaddr, netmask, &error);
	xpanelbe_proxy_destroy(daemon->priv);
	if( error != NULL)
		g_error_free(error);
	g_free(errormsg);
	if(!result)
		return FALSE;
	INFO("%s: %s -- IP='%s' NETMASK='%s'\n", __func__, netinterface, *ipaddr, *netmask);
	return TRUE;
}

gboolean blink_ports(void *context, char *type, char *value)
{
	XPanelDaemon	*daemon = context;
	DBusGProxy	*proxy;
	guchar		status = 77;
	char		*errormsg = NULL;
	GError		*error = NULL;
	gboolean	result;
	char		*type1, *value1, *res;

	INFO("%s\n", __func__);
	
	if((proxy = xpanelbe_proxy_get(daemon->priv)) == NULL)
		return FALSE;

	type1 = strdup(type);
	value1 = strdup(value);

	result = com_xorcom_XPanelBE_blink_ports (proxy, type1, value1, &status, &errormsg, &res, &error);
	xpanelbe_proxy_destroy(daemon->priv);
	INFO("%s Result=%s\n", __func__,res);
	if( error != NULL)
		g_error_free(error);
	g_free(errormsg);
	g_free(res);
	g_free(type1);
	g_free(value1);
	if(!result)
		return FALSE;
	return TRUE;
}

gboolean get_num_calls(void *context, char **num_calls)
{
	XPanelDaemon	*daemon = context;
	DBusGProxy	*proxy;
	guchar		status = 77;
	char		*errormsg = NULL;
	GError		*error = NULL;
	gboolean	result;

	INFO("%s\n", __func__);
	if((proxy = xpanelbe_proxy_get(daemon->priv)) == NULL)
		return FALSE;
	
	result = com_xorcom_XPanelBE_number_calls(proxy, &status, &errormsg, num_calls, &error);
	xpanelbe_proxy_destroy(daemon->priv);
	INFO("%s Result=%d\n", __func__,result);
	if( error != NULL)
		g_error_free(error);
	g_free(errormsg);
	if(!result)
		return FALSE;
	INFO("%s: Number of calls='%s' \n", __func__, *num_calls);
	return TRUE;
}

GList * get_rps_devices(void *context)
{
	XPanelDaemon	*daemon = context;
	DBusGProxy		*proxy;
	GList			*list = NULL;
	GError			*error = NULL;
	char			**devices = NULL;
	gboolean		result;
	int				i;

	INFO("%s\n", __func__);
	if((proxy = mps_manager_proxy_get(daemon->priv)) == NULL)
		return FALSE;

	result = com_xorcom_xrps_Manager_names (proxy, &devices, &error);
	if(!result) {
		ERR("Failed in xrps.Manager.names\n");
		return NULL;
	}
	for(i = 0; devices[i]; i++) {
		INFO("%s --> '%s'\n", __func__, devices[i]);
		list = g_list_append (list, strdup(devices[i]));
	}
	if( error != NULL)
		g_error_free(error);
	g_strfreev(devices);
	return list;
}

gboolean get_rps_mask(void *context, char * device, gint *ps_mask)
{
	XPanelDaemon	*daemon = context;
	DBusGProxy		*proxy;
	GError			*error = NULL;
	gboolean		result = FALSE;

	INFO("%s\n", __func__);
	if((proxy = mps_proxy_get(daemon->priv, device)) != NULL) {
		result = com_xorcom_xrps_get_ps_mask(proxy, ps_mask, &error);
		if(result)
			INFO("%s: ps_mask='%4.4X'\n", __func__, *ps_mask);
		if( error != NULL)
			g_error_free(error);
		mps_proxy_destroy(daemon->priv);
	}
	
	return result;
}

gboolean get_rps_voltage(void *context, char * device, int *ps_voltage)
{
	XPanelDaemon	*daemon = context;
	DBusGProxy		*proxy;
	GError			*error = NULL;
	GArray 			*ga_voltages = NULL;
	gboolean		result = FALSE;
	int				i;

	INFO("%s\n", __func__);
	if((proxy = mps_proxy_get(daemon->priv, device)) != NULL) {
		result = com_xorcom_xrps_voltage(proxy, &ga_voltages, &error);
		INFO("%s Result=%d\n", __func__,result);
		if(result) {
			INFO("%s: ps_voltage='%d','%d','%d','%d'\n", __func__, g_array_index(ga_voltages, gint, 0),
				g_array_index(ga_voltages, gint, 1), g_array_index(ga_voltages, gint, 2),
				g_array_index(ga_voltages, gint, 3));

			for(i = 0; i < 4; i++) {
				ps_voltage[i] = g_array_index(ga_voltages, gint, i);
			}
		}
		mps_proxy_destroy(daemon->priv);
		if( error != NULL)
			g_error_free(error);
		g_array_free(ga_voltages, TRUE);
	}
	return result;
}

gboolean get_twinstar_status(void *context, char **drbd_role, char **drbd_cs)
{
	XPanelDaemon	*daemon = context;
	DBusGProxy	*proxy;
	guchar		status = 77;
	char		*errormsg = NULL;
	GError		*error = NULL;
	gboolean	result;

	INFO("%s\n", __func__);
	if((proxy = xpanelbe_proxy_get(daemon->priv)) == NULL)
		return FALSE;
		
	result = com_xorcom_XPanelBE_twinstar_status(proxy, &status, &errormsg, drbd_role, drbd_cs, &error);
	xpanelbe_proxy_destroy(daemon->priv);
	if( error != NULL)
		g_error_free(error);
	g_free(errormsg);
	if(!result)
		return FALSE;
	INFO("%s: -- drbd_role='%s' drbd_cs='%s'\n", __func__,  *drbd_role, *drbd_cs);
	return TRUE;
}

gboolean get_raid_status(void *context, char **raid_sts)
{
	XPanelDaemon	*daemon = context;
	DBusGProxy	*proxy;
	guchar		status = 77;
	char		*errormsg = NULL;
	GError		*error = NULL;
	guchar		action =0;
	gboolean	result;

	INFO("%s\n", __func__);
	if((proxy = xpanelbe_proxy_get(daemon->priv)) == NULL)
		return FALSE;
		
	result = com_xorcom_XPanelBE_raid_status(proxy, &status, &errormsg, raid_sts, &action, &error);
	xpanelbe_proxy_destroy(daemon->priv);
	if( error != NULL)
		g_error_free(error);
	g_free(errormsg);
	if(!result)
		return FALSE;
	INFO("%s: -- raid_sts='%s' \n", __func__,  *raid_sts);
	return TRUE;
}

gboolean get_network_settings(void *data, struct network_settings *netsettings)
{
	XPanelDaemon		*daemon = data;
	DBusGProxy		*proxy;
	guchar			status = 77;
	char			*errormsg = NULL;
	GError			*error = NULL;
	gboolean		result;
	int			i;
	char			**net_params = NULL;

	assert(daemon);
	netsettings->raw_data = NULL;
	INFO("%s\n", __func__);
	if((proxy = xpanelbe_proxy_get(daemon->priv)) == NULL)
		return FALSE;
	result = com_xorcom_XPanelBE_network_settings (proxy, &status, &errormsg, &net_params, &error);
	xpanelbe_proxy_destroy(daemon->priv);
	INFO("%s: (%d) = status=%d errmsg='%s'\n", __func__, result, status, errormsg);
	netsettings->raw_data = net_params;
	for( i=0; net_params[ i ]; i+=2) {
	  if(strcmp(net_params[ i ], "DEFAULTGW") == 0)
		  netsettings->defaultgw = net_params[ i+1];
	  if(strcmp(net_params[ i ], "DNS1") == 0)
		  netsettings->dns1 = net_params[ i+1];
	  if(strcmp(net_params[ i ], "DNS2") == 0)
		  netsettings->dns2 = net_params[ i+1];
	}
	if( error != NULL)
		g_error_free(error);
	g_free(errormsg);
	return TRUE;
}

void free_network_settings(void *data, struct network_settings *netsettings)
{
	XPanelDaemon	*daemon = data;

	assert(daemon);
	assert(netsettings);
	fprintf(stderr, "%s\n", __func__);
	if(netsettings->raw_data)
	  	g_strfreev(netsettings->raw_data);
		
	netsettings->raw_data = NULL;
	netsettings->defaultgw = NULL;
	netsettings->dns1 = NULL;
	netsettings->dns2 = NULL;
}

gboolean apply_manualnet_update(void *context, const char *default_gtw, const char *dns1, const char *dns2,
								const char **interfaces)
{
	XPanelDaemon	*daemon = context;
	DBusGProxy	*proxy;
	guchar		status = 77;
	char		*errormsg = NULL;
	GError		*error = NULL;
	gboolean	result;

	INFO("%s\n", __func__);
	if((proxy = xpanelbe_proxy_get(daemon->priv)) == NULL)
		return FALSE;

	result = com_xorcom_XPanelBE_apply_network_settings (proxy, default_gtw, dns1, dns2, interfaces, &status, &errormsg, &error);
	xpanelbe_proxy_destroy(daemon->priv);
	if( error != NULL)
		g_error_free(error);
	g_free(errormsg);
	return result;
}

int asterisk_restart(void *context)
{
	XPanelDaemon	*daemon = context;
	DBusGProxy	*proxy;
	guchar		status = 77;
	char		*errormsg = NULL;
	GError		*error = NULL;
	gboolean	result = TRUE;

	INFO("%s\n", __func__);
	if((proxy = xpanelbe_proxy_get(daemon->priv)) == NULL)
		return -EFAULT;
	
	result = com_xorcom_XPanelBE_asterisk_restart (proxy, &status, &errormsg, &error);
	xpanelbe_proxy_destroy(daemon->priv);
	if( error != NULL)
		g_error_free(error);
	g_free(errormsg);
	return result;
}

int dahdi_restart(void *context)
{
	XPanelDaemon	*daemon = context;
	DBusGProxy	*proxy;
	guchar		status = 77;
	char		*errormsg = NULL;
	GError		*error = NULL;
	gboolean	result = TRUE;

	INFO("%s\n", __func__);
	if((proxy = xpanelbe_proxy_get(daemon->priv)) == NULL)
		return -EFAULT;
	
	result = com_xorcom_XPanelBE_dahdi_restart (proxy, &status, &errormsg, &error);
	xpanelbe_proxy_destroy(daemon->priv);
	if( error != NULL)
		g_error_free(error);
	g_free(errormsg);
	return result;
}

int server_reboot(void *context)
{
	XPanelDaemon	*daemon = context;
	DBusGProxy	*proxy;
	guchar		status = 77;
	char		*errormsg = NULL;
	GError		*error = NULL;
	gboolean	result = TRUE;

	INFO("%s\n", __func__);
	if((proxy = xpanelbe_proxy_get(daemon->priv)) == NULL)
		return -EFAULT;
	
	result = com_xorcom_XPanelBE_server_reboot (proxy, &status, &errormsg, &error);
	xpanelbe_proxy_destroy(daemon->priv);
	if( error != NULL)
		g_error_free(error);
	g_free(errormsg);
	return result;
}

int server_poweroff(void *context)
{
	XPanelDaemon	*daemon = context;
	DBusGProxy	*proxy;
	guchar		status = 77;
	char		*errormsg = NULL;
	GError		*error = NULL;
	gboolean	result = TRUE;

	INFO("%s\n", __func__);
	if((proxy = xpanelbe_proxy_get(daemon->priv)) == NULL)
		return -EFAULT;
	
	result = com_xorcom_XPanelBE_server_poweroff (proxy, &status, &errormsg, &error);
	xpanelbe_proxy_destroy(daemon->priv);
	if( error != NULL)
		g_error_free(error);
	g_free(errormsg);
	return result;
}

gboolean get_fan_speed(void *context, guchar *fan0, guchar *fan1, guchar *fan2) 
{
	XPanelDaemon	*daemon = context;
	DBusGProxy	*proxy;
	guchar		status = 77;
	GError		*error = NULL;
	char		*errormsg = NULL;
	gboolean	result = FALSE;

	INFO("%s\n", __func__);
	if((proxy = xpanelbe_proxy_get(daemon->priv)) == NULL)
		return -EFAULT;
	
	result = com_xorcom_XPanelBE_fan_speed_set (proxy, &status, &errormsg, fan0, fan1, fan2, &error);
	xpanelbe_proxy_destroy(daemon->priv);
	INFO("%s: (%d) = status=%d errmsg='%s'\n", __func__, result, status, errormsg);
	
	if (result == FALSE || status != 0)
		result = FALSE;
	else {
		result = TRUE;
		INFO("%s: fan0=%d, fan1=%d, fan2=%d\n", __func__, *fan0, *fan1, *fan2);
	}
	if( error != NULL)
		g_error_free(error);
	g_free(errormsg);
	
	return result;
}

gboolean psu_status_notification(void *context, guchar sts, guchar *action) 
{
	XPanelDaemon	*daemon = context;
	DBusGProxy	*proxy;
	guchar		status = 77;
	GError		*error = NULL;
	char		*errormsg = NULL;
	gboolean	result = FALSE;

	//INFO("%s\n", __func__);
	if((proxy = xpanelbe_proxy_get(daemon->priv)) == NULL)
		return -EFAULT;

	result = com_xorcom_XPanelBE_psu_status_notification(proxy, sts, &status, &errormsg, action, &error);
	xpanelbe_proxy_destroy(daemon->priv);
	//INFO("%s: (%d) = status=%d errmsg='%s' action=%d\n", __func__, result, status, errormsg, *action);
	
	if (result == FALSE || status != 0)
		result = FALSE;
	else
		result = TRUE;

	if( error != NULL)
		g_error_free(error);
	g_free(errormsg);
	
	return result;
}
